@@ -2,7 +2,7 @@ package capabilities_test
 
 import (
 	"github.com/cri-o/cri-o/internal/config/capabilities"
-	. "github.com/onsi/ginkgo"
+	. "github.com/onsi/ginkgo/v2"
 	. "github.com/onsi/gomega"
 )
 
