@@ -1,15 +1,15 @@
 package rdt
 
 import (
-	"io/ioutil"
+	"os"
 
-	. "github.com/onsi/ginkgo"
+	. "github.com/onsi/ginkgo/v2"
 	. "github.com/onsi/gomega"
 )
 
 func tempFileWithData(data string) string {
 	f := t.MustTempFile("")
-	Expect(ioutil.WriteFile(f, []byte(data), 0o644)).To(BeNil())
+	Expect(os.WriteFile(f, []byte(data), 0o644)).To(BeNil())
 	return f
 }
 
