@@ -7,7 +7,7 @@ import (
 	"strings"
 
 	libconfig "github.com/cri-o/cri-o/pkg/config"
-	"github.com/pkg/errors"
+	"github.com/cri-o/cri-o/server/metrics/collectors"
 	"github.com/sirupsen/logrus"
 	"github.com/urfave/cli/v2"
 )
@@ -66,6 +66,12 @@ func mergeConfig(config *libconfig.Config, ctx *cli.Context) error {
 	if err := config.UpdateFromPath(ctx.String("config-dir")); err != nil {
 		return err
 	}
+	// If "config-dir" is specified, config.UpdateFromPath() will set config.singleConfigPath as
+	// the last config file in "config-dir".
+	// We need correct it to the path specified by "config"
+	if path != "" {
+		config.SetSingleConfigPath(path)
+	}
 
 	// Override options set with the CLI.
 	if ctx.IsSet("conmon") {
@@ -130,16 +136,28 @@ func mergeConfig(config *libconfig.Config, ctx *cli.Context) error {
 			fields := strings.Split(r, ":")
 
 			runtimeType := libconfig.DefaultRuntimeType
+			privilegedWithoutHostDevices := false
+			runtimeConfigPath := ""
 
 			switch len(fields) {
+			case 6:
+				runtimeConfigPath = fields[5]
+				fallthrough
+			case 5:
+				if fields[4] == "true" {
+					privilegedWithoutHostDevices = true
+				}
+				fallthrough
 			case 4:
 				runtimeType = fields[3]
 				fallthrough
 			case 3:
 				config.Runtimes[fields[0]] = &libconfig.RuntimeHandler{
-					RuntimePath: fields[1],
-					RuntimeRoot: fields[2],
-					RuntimeType: runtimeType,
+					RuntimePath:                  fields[1],
+					RuntimeRoot:                  fields[2],
+					RuntimeType:                  runtimeType,
+					PrivilegedWithoutHostDevices: privilegedWithoutHostDevices,
+					RuntimeConfigPath:            runtimeConfigPath,
 				}
 			default:
 				return fmt.Errorf("wrong format for --runtimes: %q", r)
@@ -158,9 +176,15 @@ func mergeConfig(config *libconfig.Config, ctx *cli.Context) error {
 	if ctx.IsSet("apparmor-profile") {
 		config.ApparmorProfile = ctx.String("apparmor-profile")
 	}
+	if ctx.IsSet("blockio-config-file") {
+		config.BlockIOConfigFile = ctx.String("blockio-config-file")
+	}
 	if ctx.IsSet("irqbalance-config-file") {
 		config.IrqBalanceConfigFile = ctx.String("irqbalance-config-file")
 	}
+	if ctx.IsSet("rdt-config-file") {
+		config.RdtConfigFile = ctx.String("rdt-config-file")
+	}
 	if ctx.IsSet("cgroup-manager") {
 		config.CgroupManagerName = ctx.String("cgroup-manager")
 	}
@@ -212,9 +236,15 @@ func mergeConfig(config *libconfig.Config, ctx *cli.Context) error {
 	if ctx.IsSet("uid-mappings") {
 		config.UIDMappings = ctx.String("uid-mappings")
 	}
+	if ctx.IsSet("minimum-mappable-uid") {
+		config.MinimumMappableUID = ctx.Int64("minimum-mappable-uid")
+	}
 	if ctx.IsSet("gid-mappings") {
 		config.GIDMappings = ctx.String("gid-mappings")
 	}
+	if ctx.IsSet("minimum-mappable-gid") {
+		config.MinimumMappableGID = ctx.Int64("minimum-mappable-gid")
+	}
 	if ctx.IsSet("log-level") {
 		config.LogLevel = ctx.String("log-level")
 	}
@@ -227,6 +257,15 @@ func mergeConfig(config *libconfig.Config, ctx *cli.Context) error {
 	if ctx.IsSet("additional-devices") {
 		config.AdditionalDevices = StringSliceTrySplit(ctx, "additional-devices")
 	}
+	if ctx.IsSet("allowed-devices") {
+		config.AllowedDevices = StringSliceTrySplit(ctx, "allowed-devices")
+	}
+	if ctx.IsSet("cdi-spec-dirs") {
+		config.CDISpecDirs = StringSliceTrySplit(ctx, "cdi-spec-dirs")
+	}
+	if ctx.IsSet("device-ownership-from-security-context") {
+		config.DeviceOwnershipFromSecurityContext = ctx.Bool("device-ownership-from-security-context")
+	}
 	if ctx.IsSet("conmon-env") {
 		config.ConmonEnv = StringSliceTrySplit(ctx, "conmon-env")
 	}
@@ -239,6 +278,9 @@ func mergeConfig(config *libconfig.Config, ctx *cli.Context) error {
 	if ctx.IsSet("container-exits-dir") {
 		config.ContainerExitsDir = ctx.String("container-exits-dir")
 	}
+	if ctx.IsSet("enable-criu-support") {
+		config.EnableCriuSupport = ctx.Bool("enable-criu-support")
+	}
 	if ctx.IsSet("ctr-stop-timeout") {
 		config.CtrStopTimeout = ctx.Int64("ctr-stop-timeout")
 	}
@@ -248,9 +290,6 @@ func mergeConfig(config *libconfig.Config, ctx *cli.Context) error {
 	if ctx.IsSet("grpc-max-send-msg-size") {
 		config.GRPCMaxSendMsgSize = ctx.Int("grpc-max-send-msg-size")
 	}
-	if ctx.IsSet("manage-ns-lifecycle") {
-		config.ManageNSLifecycle = ctx.Bool("manage-ns-lifecycle")
-	}
 	if ctx.IsSet("drop-infra-ctr") {
 		config.DropInfraCtr = ctx.Bool("drop-infra-ctr")
 	}
@@ -287,6 +326,9 @@ func mergeConfig(config *libconfig.Config, ctx *cli.Context) error {
 	if ctx.IsSet("clean-shutdown-file") {
 		config.CleanShutdownFile = ctx.String("clean-shutdown-file")
 	}
+	if ctx.IsSet("absent-mount-sources-to-reject") {
+		config.AbsentMountSourcesToReject = StringSliceTrySplit(ctx, "absent-mount-sources-to-reject")
+	}
 	if ctx.IsSet("internal-wipe") {
 		config.InternalWipe = ctx.Bool("internal-wipe")
 	}
@@ -299,16 +341,36 @@ func mergeConfig(config *libconfig.Config, ctx *cli.Context) error {
 	if ctx.IsSet("metrics-socket") {
 		config.MetricsSocket = ctx.String("metrics-socket")
 	}
+	if ctx.IsSet("metrics-cert") {
+		config.MetricsCert = ctx.String("metrics-cert")
+	}
+	if ctx.IsSet("metrics-key") {
+		config.MetricsKey = ctx.String("metrics-key")
+	}
+	if ctx.IsSet("metrics-collectors") {
+		config.MetricsCollectors = collectors.FromSlice(ctx.StringSlice("metrics-collectors"))
+	}
+	if ctx.IsSet("enable-tracing") {
+		config.EnableTracing = ctx.Bool("enable-tracing")
+	}
+	if ctx.IsSet("tracing-endpoint") {
+		config.TracingEndpoint = ctx.String("tracing-endpoint")
+	}
+	if ctx.IsSet("tracing-sampling-rate-per-million") {
+		config.TracingSamplingRatePerMillion = ctx.Int("tracing-sampling-rate-per-million")
+	}
 	if ctx.IsSet("big-files-temporary-dir") {
 		config.BigFilesTemporaryDir = ctx.String("big-files-temporary-dir")
 	}
 	if ctx.IsSet("separate-pull-cgroup") {
 		config.SeparatePullCgroup = ctx.String("separate-pull-cgroup")
 	}
-
 	if ctx.IsSet("infra-ctr-cpuset") {
 		config.InfraCtrCPUSet = ctx.String("infra-ctr-cpuset")
 	}
+	if ctx.IsSet("stats-collection-period") {
+		config.StatsCollectionPeriod = ctx.Int("stats-collection-period")
+	}
 
 	return nil
 }
@@ -316,7 +378,7 @@ func mergeConfig(config *libconfig.Config, ctx *cli.Context) error {
 func GetFlagsAndMetadata() ([]cli.Flag, map[string]interface{}, error) {
 	config, err := libconfig.DefaultConfig()
 	if err != nil {
-		return nil, nil, errors.Errorf("error loading server config: %v", err)
+		return nil, nil, fmt.Errorf("error loading server config: %w", err)
 	}
 
 	// TODO FIXME should be crio wipe flags
@@ -357,13 +419,13 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 		},
 		&cli.StringFlag{
 			Name:      "conmon",
-			Usage:     fmt.Sprintf("Path to the conmon binary, used for monitoring the OCI runtime. Will be searched for using $PATH if empty. (default: %q)", defConf.Conmon),
+			Usage:     fmt.Sprintf("Path to the conmon binary, used for monitoring the OCI runtime. Will be searched for using $PATH if empty. This option is deprecated, and will be removed in the future. (default: %q)", defConf.Conmon),
 			EnvVars:   []string{"CONTAINER_CONMON"},
 			TakesFile: true,
 		},
 		&cli.StringFlag{
 			Name:    "conmon-cgroup",
-			Usage:   "cgroup to be used for conmon process",
+			Usage:   "cgroup to be used for conmon process. This option is deprecated and will be removed in the future.",
 			Value:   defConf.ConmonCgroup,
 			EnvVars: []string{"CONTAINER_CONMON_CGROUP"},
 		},
@@ -519,7 +581,7 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 		},
 		&cli.StringSliceFlag{
 			Name:    "runtimes",
-			Usage:   "OCI runtimes, format is runtime_name:runtime_path:runtime_root:runtime_type",
+			Usage:   "OCI runtimes, format is runtime_name:runtime_path:runtime_root:runtime_type:privileged_without_host_devices:runtime_config_path",
 			EnvVars: []string{"CONTAINER_RUNTIMES"},
 		},
 		&cli.StringFlag{
@@ -528,10 +590,11 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 			EnvVars:   []string{"CONTAINER_SECCOMP_PROFILE"},
 			TakesFile: true,
 		},
-		&cli.StringFlag{
+		&cli.BoolFlag{
 			Name:    "seccomp-use-default-when-empty",
-			Usage:   fmt.Sprintf("Use the default seccomp profile when an empty one is specified (default: %t)", defConf.SeccompUseDefaultWhenEmpty),
+			Usage:   "Use the default seccomp profile when an empty one is specified",
 			EnvVars: []string{"CONTAINER_SECCOMP_USE_DEFAULT_WHEN_EMPTY"},
+			Value:   defConf.Seccomp().UseDefaultWhenEmpty(),
 		},
 		&cli.StringFlag{
 			Name:    "apparmor-profile",
@@ -539,11 +602,21 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 			Value:   defConf.ApparmorProfile,
 			EnvVars: []string{"CONTAINER_APPARMOR_PROFILE"},
 		},
+		&cli.StringFlag{
+			Name:  "blockio-config-file",
+			Usage: "Path to the blockio class configuration file for configuring the cgroup blockio controller.",
+			Value: defConf.BlockIOConfigFile,
+		},
 		&cli.StringFlag{
 			Name:  "irqbalance-config-file",
 			Usage: "The irqbalance service config file which is used by CRI-O.",
 			Value: defConf.IrqBalanceConfigFile,
 		},
+		&cli.StringFlag{
+			Name:  "rdt-config-file",
+			Usage: "Path to the RDT configuration file for configuring the resctrl pseudo-filesystem",
+			Value: defConf.RdtConfigFile,
+		},
 		&cli.BoolFlag{
 			Name:    "selinux",
 			Usage:   fmt.Sprintf("Enable selinux support (default: %t)", defConf.SELinux),
@@ -558,13 +631,13 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 		&cli.Int64Flag{
 			Name:    "pids-limit",
 			Value:   libconfig.DefaultPidsLimit,
-			Usage:   "Maximum number of processes allowed in a container",
+			Usage:   "Maximum number of processes allowed in a container. This option is deprecated. The Kubelet flag '--pod-pids-limit' should be used instead.",
 			EnvVars: []string{"CONTAINER_PIDS_LIMIT"},
 		},
 		&cli.Int64Flag{
 			Name:    "log-size-max",
 			Value:   libconfig.DefaultLogSizeMax,
-			Usage:   "Maximum log size in bytes for a container. If it is positive, it must be >= 8192 to match/exceed conmon read buffer",
+			Usage:   "Maximum log size in bytes for a container. If it is positive, it must be >= 8192 to match/exceed conmon read buffer. This option is deprecated. The Kubelet flag '--container-log-max-size' should be used instead.",
 			EnvVars: []string{"CONTAINER_LOG_SIZE_MAX"},
 		},
 		&cli.BoolFlag{
@@ -654,6 +727,16 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 			Usage:   "Enable pprof remote profiler on localhost:6060",
 			EnvVars: []string{"CONTAINER_PROFILE"},
 		},
+		&cli.StringFlag{
+			Name:    "profile-cpu",
+			Usage:   "Write a pprof CPU profile to the provided path",
+			EnvVars: []string{"CONTAINER_PROFILE_CPU"},
+		},
+		&cli.StringFlag{
+			Name:    "profile-mem",
+			Usage:   "Write a pprof memory profile to the provided path",
+			EnvVars: []string{"CONTAINER_PROFILE_MEM"},
+		},
 		&cli.IntFlag{
 			Name:    "profile-port",
 			Value:   6060,
@@ -681,6 +764,39 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 			Usage:   "Socket for the metrics endpoint",
 			EnvVars: []string{"CONTAINER_METRICS_SOCKET"},
 		},
+		&cli.StringFlag{
+			Name:    "metrics-cert",
+			Usage:   "Certificate for the secure metrics endpoint",
+			EnvVars: []string{"CONTAINER_METRICS_CERT"},
+		},
+		&cli.StringFlag{
+			Name:    "metrics-key",
+			Usage:   "Certificate key for the secure metrics endpoint",
+			EnvVars: []string{"CONTAINER_METRICS_KEY"},
+		},
+		&cli.StringSliceFlag{
+			Name:    "metrics-collectors",
+			Usage:   "Enabled metrics collectors",
+			Value:   cli.NewStringSlice(collectors.All().ToSlice()...),
+			EnvVars: []string{"CONTAINER_METRICS_COLLECTORS"},
+		},
+		&cli.BoolFlag{
+			Name:    "enable-tracing",
+			Usage:   "Enable OpenTelemetry trace data exporting",
+			EnvVars: []string{"CONTAINER_ENABLE_TRACING"},
+		},
+		&cli.IntFlag{
+			Name:    "tracing-sampling-rate-per-million",
+			Value:   defConf.TracingSamplingRatePerMillion,
+			Usage:   "Number of samples to collect per million OpenTelemetry spans",
+			EnvVars: []string{"CONTAINER_TRACING_SAMPLING_RATE_PER_MILLION"},
+		},
+		&cli.StringFlag{
+			Name:    "tracing-endpoint",
+			Value:   defConf.TracingEndpoint,
+			Usage:   "Address on which the gRPC tracing collector will listen",
+			EnvVars: []string{"CONTAINER_TRACING_ENDPOINT"},
+		},
 		&cli.StringFlag{
 			Name:    "big-files-temporary-dir",
 			Usage:   `Path to the temporary directory to use for storing big files, used to store image blobs and data streams related to containers image management.`,
@@ -708,16 +824,44 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 			Value:   "",
 			EnvVars: []string{"CONTAINER_GID_MAPPINGS"},
 		},
+		&cli.Int64Flag{
+			Name:    "minimum-mappable-uid",
+			Usage:   "Specify the lowest host UID which can be specified in mappings for a pod that will be run as a UID other than 0",
+			Value:   defConf.MinimumMappableUID,
+			EnvVars: []string{"CONTAINER_MINIMUM_MAPPABLE_UID"},
+		},
+		&cli.Int64Flag{
+			Name:    "minimum-mappable-gid",
+			Usage:   "Specify the lowest host GID which can be specified in mappings for a pod that will be run as a UID other than 0",
+			Value:   defConf.MinimumMappableGID,
+			EnvVars: []string{"CONTAINER_MINIMUM_MAPPABLE_GID"},
+		},
+		&cli.StringSliceFlag{
+			Name:    "allowed-devices",
+			Usage:   "Devices a user is allowed to specify with the \"io.kubernetes.cri-o.Devices\" allowed annotation",
+			Value:   cli.NewStringSlice(defConf.AllowedDevices...),
+			EnvVars: []string{"CONTAINER_ALLOWED_DEVICES"},
+		},
 		&cli.StringSliceFlag{
 			Name:    "additional-devices",
 			Usage:   "Devices to add to the containers ",
 			Value:   cli.NewStringSlice(defConf.AdditionalDevices...),
 			EnvVars: []string{"CONTAINER_ADDITIONAL_DEVICES"},
 		},
+		&cli.StringSliceFlag{
+			Name:    "cdi-spec-dirs",
+			Usage:   "Directories to scan for CDI Spec files",
+			Value:   cli.NewStringSlice(defConf.CDISpecDirs...),
+			EnvVars: []string{"CONTAINER_CDI_SPEC_DIRS"},
+		},
+		&cli.BoolFlag{
+			Name:  "device-ownership-from-security-context",
+			Usage: "Set devices' uid/gid ownership from runAsUser/runAsGroup",
+		},
 		&cli.StringSliceFlag{
 			Name:    "conmon-env",
 			Value:   cli.NewStringSlice(defConf.ConmonEnv...),
-			Usage:   "Environment variable list for the conmon process, used for passing necessary environment variables to conmon or the runtime",
+			Usage:   "Environment variable list for the conmon process, used for passing necessary environment variables to conmon or the runtime. This option is deprecated and will be removed in the future.",
 			EnvVars: []string{"CONTAINER_CONMON_ENV"},
 		},
 		&cli.StringSliceFlag{
@@ -757,14 +901,9 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 			Value:   defConf.GRPCMaxSendMsgSize,
 			EnvVars: []string{"CONTAINER_GRPC_MAX_SEND_MSG_SIZE"},
 		},
-		&cli.BoolFlag{
-			Name:    "manage-ns-lifecycle",
-			Usage:   fmt.Sprintf("Determines whether we pin and remove IPC, network and UTS namespaces and manage their lifecycle. This option is being deprecated, and will be unconditionally true in the future. (default: %v)", defConf.ManageNSLifecycle),
-			EnvVars: []string{"CONTAINER_MANAGE_NS_LIFECYCLE"},
-		},
 		&cli.BoolFlag{
 			Name:    "drop-infra-ctr",
-			Usage:   fmt.Sprintf("Determines whether pods are created without an infra container (when the pod is not using a pod level PID namespace). Requires ManageNSLifecycle to be true (default: %v)", defConf.DropInfraCtr),
+			Usage:   fmt.Sprintf("Determines whether pods are created without an infra container, when the pod is not using a pod level PID namespace (default: %v)", defConf.DropInfraCtr),
 			EnvVars: []string{"CONTAINER_DROP_INFRA_CTR"},
 		},
 		&cli.StringFlag{
@@ -819,6 +958,19 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 			EnvVars:     []string{"CONTAINER_REGISTRIES_CONF"},
 			TakesFile:   true,
 		},
+		&cli.StringFlag{
+			Name:        "registries-conf-dir",
+			Usage:       "path to the registries.conf.d directory",
+			Destination: &defConf.SystemContext.SystemRegistriesConfDirPath,
+			Hidden:      true,
+			EnvVars:     []string{"CONTAINER_REGISTRIES_CONF_DIR"},
+			TakesFile:   true,
+		},
+		&cli.StringFlag{
+			Name:   "address",
+			Usage:  "address used for the publish command",
+			Hidden: true,
+		},
 		&cli.StringFlag{
 			Name:      "version-file",
 			Usage:     "Location for CRI-O to lay down the temporary version file. It is used to check if crio wipe should wipe containers, which should always happen on a node reboot",
@@ -835,7 +987,7 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 		},
 		&cli.BoolFlag{
 			Name:    "internal-wipe",
-			Usage:   "Whether CRI-O should wipe containers after a reboot and images after an upgrade when the server starts. If set to false, one must run `crio wipe` to wipe the containers and images in these situations.",
+			Usage:   "Whether CRI-O should wipe containers after a reboot and images after an upgrade when the server starts. If set to false, one must run `crio wipe` to wipe the containers and images in these situations. This option is deprecated, and will be removed in the future.",
 			Value:   defConf.InternalWipe,
 			EnvVars: []string{"CONTAINER_INTERNAL_WIPE"},
 		},
@@ -851,6 +1003,24 @@ func getCrioFlags(defConf *libconfig.Config) []cli.Flag {
 			EnvVars:   []string{"CONTAINER_CLEAN_SHUTDOWN_FILE"},
 			TakesFile: true,
 		},
+		&cli.StringSliceFlag{
+			Name:    "absent-mount-sources-to-reject",
+			Value:   cli.NewStringSlice(defConf.AbsentMountSourcesToReject...),
+			Usage:   "A list of paths that, when absent from the host, will cause a container creation to fail (as opposed to the current behavior of creating a directory).",
+			EnvVars: []string{"CONTAINER_ABSENT_MOUNT_SOURCES_TO_REJECT"},
+		},
+		&cli.IntFlag{
+			Name:    "stats-collection-period",
+			Value:   defConf.StatsCollectionPeriod,
+			Usage:   "The number of seconds between collecting pod and container stats. If set to 0, the stats are collected on-demand instead.",
+			EnvVars: []string{"CONTAINER_STATS_COLLECTION_PERIOD"},
+		},
+		&cli.BoolFlag{
+			Name:    "enable-criu-support",
+			Usage:   "Enable CRIU integration, requires that the criu binary is available in $PATH. (default: '')",
+			EnvVars: []string{"CONTAINER_ENABLE_CRIU_SUPPORT"},
+			Value:   false,
+		},
 	}
 }
 
