@@ -4,8 +4,8 @@ import (
 	"context"
 	"crypto/tls"
 	"crypto/x509"
+	"errors"
 	"fmt"
-	"io/ioutil"
 	"net"
 	"net/http"
 	"os"
@@ -29,19 +29,16 @@ import (
 	"github.com/cri-o/cri-o/internal/storage"
 	"github.com/cri-o/cri-o/internal/version"
 	libconfig "github.com/cri-o/cri-o/pkg/config"
-	"github.com/cri-o/cri-o/server/cri/types"
 	"github.com/cri-o/cri-o/server/metrics"
 	"github.com/cri-o/cri-o/server/streaming"
 	"github.com/cri-o/cri-o/utils"
 	"github.com/fsnotify/fsnotify"
-	"github.com/pkg/errors"
-	"github.com/prometheus/client_golang/prometheus/promhttp"
 	"github.com/sirupsen/logrus"
 	"golang.org/x/sys/unix"
+	types "k8s.io/cri-api/pkg/apis/runtime/v1"
 )
 
 const (
-	shutdownFile        = "/var/lib/crio/crio.shutdown"
 	certRefreshInterval = time.Minute * 5
 	rootlessEnvName     = "_CRIO_ROOTLESS"
 )
@@ -67,7 +64,7 @@ type Server struct {
 	monitorsChan      chan struct{}
 	defaultIDMappings *idtools.IDMappings
 
-	updateLock sync.RWMutex
+	minimumMappableUID, minimumMappableGID int64
 
 	// pullOperationsInProgress is used to avoid pulling the same image in parallel. Goroutines
 	// will block on the pullResult.
@@ -122,9 +119,9 @@ func (cc *certConfigCache) GetConfigForClient(hello *tls.ClientHelloInfo) (*tls.
 	}
 	config.Certificates = []tls.Certificate{cert}
 	if len(cc.tlsCA) > 0 {
-		caBytes, err := ioutil.ReadFile(cc.tlsCA)
+		caBytes, err := os.ReadFile(cc.tlsCA)
 		if err != nil {
-			return nil, errors.Wrap(err, "read TLS CA file")
+			return nil, fmt.Errorf("read TLS CA file: %w", err)
 		}
 		certPool := x509.NewCertPool()
 		certPool.AppendCertsFromPEM(caBytes)
@@ -169,7 +166,7 @@ func (s *Server) restore(ctx context.Context) []string {
 	containersAndTheirImages := map[string]string{}
 	containers, err := s.Store().Containers()
 	if err != nil && !errors.Is(err, os.ErrNotExist) {
-		log.Warnf(ctx, "could not read containers and sandboxes: %v", err)
+		log.Warnf(ctx, "Could not read containers and sandboxes: %v", err)
 	}
 	pods := map[string]*storage.RuntimeContainerMetadata{}
 	podContainers := map[string]*storage.RuntimeContainerMetadata{}
@@ -178,11 +175,11 @@ func (s *Server) restore(ctx context.Context) []string {
 	for i := range containers {
 		metadata, err2 := s.StorageRuntimeServer().GetContainerMetadata(containers[i].ID)
 		if err2 != nil {
-			log.Warnf(ctx, "error parsing metadata for %s: %v, ignoring", containers[i].ID, err2)
+			log.Warnf(ctx, "Error parsing metadata for %s: %v, ignoring", containers[i].ID, err2)
 			continue
 		}
 		if !storage.IsCrioContainer(&metadata) {
-			log.Debugf(ctx, "container %s determined to not be a CRI-O container or sandbox", containers[i].ID)
+			log.Debugf(ctx, "Container %s determined to not be a CRI-O container or sandbox", containers[i].ID)
 			continue
 		}
 		names[containers[i].ID] = containers[i].Names
@@ -201,26 +198,27 @@ func (s *Server) restore(ctx context.Context) []string {
 		if err == nil {
 			continue
 		}
-		log.Warnf(ctx, "could not restore sandbox %s; deleting it and containers underneath it: %v", sbID, err)
+		log.Warnf(ctx, "Could not restore sandbox %s: %v", sbID, err)
 		for _, n := range names[sbID] {
 			if err := s.Store().DeleteContainer(n); err != nil && err != storageTypes.ErrNotAContainer {
-				log.Warnf(ctx, "unable to delete container %s: %v", n, err)
+				log.Warnf(ctx, "Unable to delete container %s: %v", n, err)
 			}
 			// Release the infra container name and the pod name for future use
-			if strings.Contains(n, types.InfraContainerName) {
+			if strings.Contains(n, oci.InfraContainerName) {
 				s.ReleaseContainerName(n)
 			} else {
 				s.ReleasePodName(n)
 			}
 		}
 		// Go through the containers and delete any container that was under the deleted pod
+		log.Warnf(ctx, "Deleting all containers under sandbox %s since it could not be restored", sbID)
 		for k, v := range podContainers {
 			if v.PodID != sbID {
 				continue
 			}
 			for _, n := range names[k] {
 				if err := s.Store().DeleteContainer(n); err != nil && err != storageTypes.ErrNotAContainer {
-					log.Warnf(ctx, "unable to delete container %s: %v", n, err)
+					log.Warnf(ctx, "Unable to delete container %s: %v", n, err)
 				}
 				// Release the container name for future use
 				s.ReleaseContainerName(n)
@@ -295,35 +293,19 @@ func (s *Server) restore(ctx context.Context) []string {
 	return imagesOfDeletedContainers
 }
 
-// cleanupSandboxesOnShutdown Remove all running Sandboxes on system shutdown
-func (s *Server) cleanupSandboxesOnShutdown(ctx context.Context) {
-	_, err := os.Stat(shutdownFile)
-	if err == nil || !os.IsNotExist(err) {
-		log.Debugf(ctx, "shutting down all sandboxes, on shutdown")
-		s.stopAllPodSandboxes(ctx)
-		err = os.Remove(shutdownFile)
-		if err != nil {
-			log.Warnf(ctx, "Failed to remove %q", shutdownFile)
-		}
-	}
-}
-
 // Shutdown attempts to shut down the server's storage cleanly
 func (s *Server) Shutdown(ctx context.Context) error {
-	// why do this on clean shutdown! we want containers left running when crio
-	// is down for whatever reason no?!
-	// notice this won't trigger just on system halt but also on normal
-	// crio.service restart!!!
-	s.cleanupSandboxesOnShutdown(ctx)
+	s.config.CNIManagerShutdown()
 	s.resourceStore.Close()
 
 	if err := s.ContainerServer.Shutdown(); err != nil {
 		return err
 	}
 
-	// first, make sure we sync all storage changes
-	if err := utils.Sync(s.Store().GraphRoot()); err != nil {
-		return errors.Wrapf(err, "failed to sync graph root after shutting down")
+	// first, make sure we sync all the changes to the file system holding
+	// the graph root
+	if err := utils.Syncfs(s.Store().GraphRoot()); err != nil {
+		return fmt.Errorf("failed to sync graph root after shutting down: %w", err)
 	}
 
 	if s.config.CleanShutdownFile != "" {
@@ -334,7 +316,7 @@ func (s *Server) Shutdown(ctx context.Context) error {
 		// is not so.
 		f, err := os.Create(s.config.CleanShutdownFile)
 		if err != nil {
-			return errors.Wrapf(err, "failed to write file to indicate a clean shutdown")
+			return fmt.Errorf("failed to write file to indicate a clean shutdown: %w", err)
 		}
 		f.Close()
 
@@ -344,7 +326,7 @@ func (s *Server) Shutdown(ctx context.Context) error {
 		// However, that's much less likely than if we don't have a second Sync,
 		// and less risky than if we don't Sync after the Create
 		if err := utils.SyncParent(s.config.CleanShutdownFile); err != nil {
-			return errors.Wrapf(err, "failed to sync clean shutdown file")
+			return fmt.Errorf("failed to sync clean shutdown file: %w", err)
 		}
 	}
 
@@ -354,9 +336,9 @@ func (s *Server) Shutdown(ctx context.Context) error {
 // configureMaxThreads sets the Go runtime max threads threshold
 // which is 90% of the kernel setting from /proc/sys/kernel/threads-max
 func configureMaxThreads() error {
-	mt, err := ioutil.ReadFile("/proc/sys/kernel/threads-max")
+	mt, err := os.ReadFile("/proc/sys/kernel/threads-max")
 	if err != nil {
-		return errors.Wrap(err, "read max threads file")
+		return fmt.Errorf("read max threads file: %w", err)
 	}
 	mtint, err := strconv.Atoi(strings.TrimSpace(string(mt)))
 	if err != nil {
@@ -395,6 +377,8 @@ func New(
 	}
 	config := configIface.GetData()
 
+	useDefaultUmask()
+
 	config.SystemContext.AuthFilePath = config.GlobalAuthFile
 	config.SystemContext.SignaturePolicyPath = config.SignaturePolicyPath
 
@@ -435,6 +419,8 @@ func New(
 		config:                   *config,
 		monitorsChan:             make(chan struct{}),
 		defaultIDMappings:        idMappings,
+		minimumMappableUID:       config.MinimumMappableUID,
+		minimumMappableGID:       config.MinimumMappableGID,
 		pullOperationsInProgress: make(map[pullArguments]*pullOperation),
 		resourceStore:            resourcestore.New(),
 	}
@@ -446,16 +432,15 @@ func New(
 	// Close stdin, so shortnames will not prompt
 	devNullFile, err := os.Open(os.DevNull)
 	if err != nil {
-		return nil, errors.Wrap(err, "open devnull file")
+		return nil, fmt.Errorf("open devnull file: %w", err)
 	}
 
 	defer devNullFile.Close()
 	if err := unix.Dup2(int(devNullFile.Fd()), int(os.Stdin.Fd())); err != nil {
-		return nil, errors.Wrap(err, "close stdin")
+		return nil, fmt.Errorf("close stdin: %w", err)
 	}
 
 	deletedImages := s.restore(ctx)
-	s.cleanupSandboxesOnShutdown(ctx)
 	s.wipeIfAppropriate(ctx, deletedImages)
 
 	var bindAddressStr string
@@ -500,7 +485,7 @@ func New(
 	}
 	s.stream.ctx = ctx
 	s.stream.runtimeServer = s
-	s.stream.streamServer, err = streaming.NewServer(ctx, streamServerConfig, s.stream)
+	s.stream.streamServer, err = streaming.NewServer(ctx, &streamServerConfig, s.stream)
 	if err != nil {
 		return nil, fmt.Errorf("unable to create streaming server")
 	}
@@ -513,19 +498,34 @@ func New(
 		}
 	}()
 
-	log.Debugf(ctx, "sandboxes: %v", s.ContainerServer.ListSandboxes())
+	log.Debugf(ctx, "Sandboxes: %v", s.ContainerServer.ListSandboxes())
 
 	// Start a configuration watcher for the default config
 	s.config.StartWatcher()
 
 	// Start the metrics server if configured to be enabled
-	if err := s.startMetricsServer(); err != nil {
-		return nil, err
+	if s.config.EnableMetrics {
+		if err := metrics.New(&s.config.MetricsConfig).Start(s.monitorsChan); err != nil {
+			return nil, err
+		}
+	} else {
+		logrus.Debug("Metrics are disabled")
 	}
 
 	return s, nil
 }
 
+func useDefaultUmask() {
+	const defaultUmask = 0o022
+	oldUmask := unix.Umask(defaultUmask)
+	if oldUmask != defaultUmask {
+		logrus.Infof(
+			"Using default umask 0o%#o instead of 0o%#o",
+			defaultUmask, oldUmask,
+		)
+	}
+}
+
 // wipeIfAppropriate takes a list of images. If the config's VersionFilePersist
 // indicates an upgrade has happened, it attempts to wipe that list of images.
 // This attempt is best-effort.
@@ -533,20 +533,58 @@ func (s *Server) wipeIfAppropriate(ctx context.Context, imagesToDelete []string)
 	if !s.config.InternalWipe {
 		return
 	}
+	var (
+		shouldWipeContainers, shouldWipeImages bool
+		err                                    error
+	)
+
 	// Check if our persistent version file is out of date.
 	// If so, we have upgrade, and we should wipe images.
-	shouldWipeImages, err := version.ShouldCrioWipe(s.config.VersionFilePersist)
+	shouldWipeImages, err = version.ShouldCrioWipe(s.config.VersionFilePersist)
 	if err != nil {
-		log.Warnf(ctx, "error encountered when checking whether cri-o should wipe images: %v", err)
+		log.Warnf(ctx, "Error encountered when checking whether cri-o should wipe images: %v", err)
+	}
+
+	// Unconditionally wipe containers if we should wipe images.
+	if shouldWipeImages {
+		shouldWipeContainers = true
+	} else {
+		// Check if our version file is out of date.
+		// If so, we rebooted, and we should wipe containers.
+		shouldWipeContainers, err = version.ShouldCrioWipe(s.config.VersionFile)
+		if err != nil {
+			log.Warnf(ctx, "Error encountered when checking whether cri-o should wipe containers: %v", err)
+		}
+	}
+
+	// Translate to a map so the images are only attempted to be deleted once.
+	imageMapToDelete := make(map[string]struct{})
+	for _, img := range imagesToDelete {
+		imageMapToDelete[img] = struct{}{}
+	}
+
+	// Attempt to wipe containers, adding the images that were removed on the way.
+	if shouldWipeContainers {
+		// Best-effort append to imageMapToDelete
+		if ctrs, err := s.ContainerServer.ListContainers(); err == nil {
+			for _, ctr := range ctrs {
+				imageMapToDelete[ctr.ImageRef()] = struct{}{}
+			}
+		}
+		for _, sb := range s.ContainerServer.ListSandboxes() {
+			if err := s.removePodSandbox(ctx, sb); err != nil {
+				log.Warnf(ctx, "Failed to remove sandbox %s: %v", sb.ID(), err)
+			}
+		}
 	}
 
 	// Note: some of these will fail if some aspect of the pod cleanup failed as well,
 	// but this is best-effort anyway, as the Kubelet will eventually cleanup images when
 	// disk usage gets too high.
 	if shouldWipeImages {
-		for _, img := range imagesToDelete {
+		for img := range imageMapToDelete {
 			if err := s.removeImage(ctx, img); err != nil {
-				log.Warnf(ctx, "failed to remove image %s: %v", img, err)
+				log.Warnf(ctx, "Failed to remove image %s: %v", img, err)
 			}
 		}
 	}
@@ -591,7 +629,7 @@ func (s *Server) getPodSandboxFromRequest(podSandboxID string) (*sandbox.Sandbox
 
 	sandboxID, err := s.PodIDIndex().Get(podSandboxID)
 	if err != nil {
-		return nil, fmt.Errorf("PodSandbox with ID starting with %s not found: %v", podSandboxID, err)
+		return nil, fmt.Errorf("PodSandbox with ID starting with %s not found: %w", podSandboxID, err)
 	}
 
 	sb := s.getSandbox(sandboxID)
@@ -604,62 +642,6 @@ func (s *Server) getPodSandboxFromRequest(podSandboxID string) (*sandbox.Sandbox
 	return sb, nil
 }
 
-func (s *Server) startMetricsServer() error {
-	if !s.config.EnableMetrics {
-		return nil
-	}
-
-	me, err := s.CreateMetricsEndpoint()
-	if err != nil {
-		return errors.Wrap(err, "failed to create metrics endpoint")
-	}
-
-	if err := startMetricsEndpoint(
-		"tcp", fmt.Sprintf(":%v", s.config.MetricsPort), me,
-	); err != nil {
-		return errors.Wrap(err, "creating tcp metrics endpoint")
-	}
-
-	metricsSocket := s.config.MetricsSocket
-	if metricsSocket != "" {
-		if err := libconfig.RemoveUnusedSocket(metricsSocket); err != nil {
-			return errors.Wrapf(err, "removing ununsed socket %s", metricsSocket)
-		}
-
-		return errors.Wrap(
-			startMetricsEndpoint("unix", s.config.MetricsSocket, me),
-			"creating path metrics endpoint",
-		)
-	}
-
-	return nil
-}
-
-func startMetricsEndpoint(network, address string, me *http.ServeMux) error {
-	l, err := net.Listen(network, address)
-	if err != nil {
-		return errors.Wrap(err, "creating listener")
-	}
-
-	go func() {
-		logrus.Infof("Serving metrics on %s", address)
-		if err := http.Serve(l, me); err != nil {
-			logrus.Fatalf("failed to serve metrics endpoint %v: %v", l, err)
-		}
-	}()
-
-	return nil
-}
-
-// CreateMetricsEndpoint creates a /metrics endpoint
-// for prometheus monitoring
-func (s *Server) CreateMetricsEndpoint() (*http.ServeMux, error) {
-	metrics.Register()
-	mux := &http.ServeMux{}
-	mux.Handle("/metrics", promhttp.Handler())
-	return mux, nil
-}
-
 // StopMonitors stops all the monitors
 func (s *Server) StopMonitors() {
 	close(s.monitorsChan)
@@ -678,57 +660,59 @@ func (s *Server) StartExitMonitor(ctx context.Context) {
 		log.Fatalf(ctx, "Failed to create new watch: %v", err)
 	}
 	defer watcher.Close()
-
 	done := make(chan struct{})
-	go func() {
-		for {
-			select {
-			case event := <-watcher.Events:
-				log.Debugf(ctx, "event: %v", event)
-				if event.Op&fsnotify.Create == fsnotify.Create {
-					containerID := filepath.Base(event.Name)
-					log.Debugf(ctx, "container or sandbox exited: %v", containerID)
-					c := s.GetContainer(containerID)
-					if c != nil {
-						log.Debugf(ctx, "container exited and found: %v", containerID)
-						err := s.Runtime().UpdateContainerStatus(ctx, c)
-						if err != nil {
-							log.Warnf(ctx, "Failed to update container status %s: %v", containerID, err)
-						} else if err := s.ContainerStateToDisk(ctx, c); err != nil {
-							log.Warnf(ctx, "unable to write containers %s state to disk: %v", c.ID(), err)
-						}
-					} else {
-						sb := s.GetSandbox(containerID)
-						if sb != nil {
-							c := sb.InfraContainer()
-							if c == nil {
-								log.Warnf(ctx, "no infra container set for sandbox: %v", containerID)
-								continue
-							}
-							log.Debugf(ctx, "sandbox exited and found: %v", containerID)
-							err := s.Runtime().UpdateContainerStatus(ctx, c)
-							if err != nil {
-								log.Warnf(ctx, "Failed to update sandbox infra container status %s: %v", c.ID(), err)
-							} else if err := s.ContainerStateToDisk(ctx, c); err != nil {
-								log.Warnf(ctx, "unable to write containers %s state to disk: %v", c.ID(), err)
-							}
-						}
-					}
-				}
-			case err := <-watcher.Errors:
-				log.Debugf(ctx, "watch error: %v", err)
-				close(done)
-				return
-			case <-s.monitorsChan:
-				log.Debugf(ctx, "closing exit monitor...")
-				close(done)
-				return
-			}
-		}
-	}()
+	go s.monitorExits(ctx, watcher, done)
+
 	if err := watcher.Add(s.config.ContainerExitsDir); err != nil {
-		log.Errorf(ctx, "watcher.Add(%q) failed: %s", s.config.ContainerExitsDir, err)
+		log.Errorf(ctx, "Watcher.Add(%q) failed: %s", s.config.ContainerExitsDir, err)
 		close(done)
 	}
 	<-done
 }
+
+func (s *Server) monitorExits(ctx context.Context, watcher *fsnotify.Watcher, done chan struct{}) {
+	for {
+		select {
+		case event := <-watcher.Events:
+			go s.handleExit(ctx, event)
+		case err := <-watcher.Errors:
+			log.Debugf(ctx, "Watch error: %v", err)
+			close(done)
+			return
+		case <-s.monitorsChan:
+			log.Debugf(ctx, "Closing exit monitor...")
+			close(done)
+			return
+		}
+	}
+}
+
+func (s *Server) handleExit(ctx context.Context, event fsnotify.Event) {
+	log.Debugf(ctx, "Event: %v", event)
+	if event.Op&fsnotify.Create != fsnotify.Create {
+		return
+	}
+	containerID := filepath.Base(event.Name)
+	log.Debugf(ctx, "Container or sandbox exited: %v", containerID)
+	c := s.GetContainer(containerID)
+	resource := "container"
+	if c == nil {
+		sb := s.GetSandbox(containerID)
+		if sb == nil {
+			return
+		}
+		c = sb.InfraContainer()
+		resource = "sandbox infra"
+	}
+	log.Debugf(ctx, "%s exited and found: %v", resource, containerID)
+	if err := s.Runtime().UpdateContainerStatus(ctx, c); err != nil {
+		log.Warnf(ctx, "Failed to update %s status %s: %v", resource, containerID, err)
+		return
+	}
+	if err := s.ContainerStateToDisk(ctx, c); err != nil {
+		log.Warnf(ctx, "Unable to write %s %s state to disk: %v", resource, c.ID(), err)
+	}
+	if err := os.Remove(event.Name); err != nil {
+		log.Warnf(ctx, "Failed to remove exit file: %v", err)
+	}
+}
