@@ -1,9 +1,11 @@
 package server
 
 import (
+	"context"
+	"fmt"
+
 	"github.com/cri-o/cri-o/internal/version"
-	"golang.org/x/net/context"
-	pb "k8s.io/cri-api/pkg/apis/runtime/v1alpha2"
+	types "k8s.io/cri-api/pkg/apis/runtime/v1"
 )
 
 const (
@@ -12,16 +14,19 @@ const (
 	kubeAPIVersion = "0.1.0"
 	// containerName is the name prepended in kubectl describe->Container ID:
 	// cri-o://<CONTAINER_ID>
-	containerName     = "cri-o"
-	runtimeAPIVersion = "v1alpha1"
+	containerName = "cri-o"
 )
 
 // Version returns the runtime name, runtime version and runtime API version
-func (s *Server) Version(ctx context.Context, req *pb.VersionRequest) (*pb.VersionResponse, error) {
-	return &pb.VersionResponse{
+func (s *Server) Version(_ context.Context, apiVersion string) (*types.VersionResponse, error) {
+	info, err := version.Get(false)
+	if err != nil {
+		return nil, fmt.Errorf("get server version: %w", err)
+	}
+	return &types.VersionResponse{
 		Version:           kubeAPIVersion,
 		RuntimeName:       containerName,
-		RuntimeVersion:    version.Get().Version,
-		RuntimeApiVersion: runtimeAPIVersion,
+		RuntimeVersion:    info.Version,
+		RuntimeApiVersion: apiVersion,
 	}, nil
 }
